unit Fractal2Unit;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, ComplexUnit;

type
  TDoubleRect = record
    Left, Top, Right, Bottom: Double;
  end;

  TViewChangeProcedure = procedure (NewView: TDoubleRect) of object;
  TProgressProcedure = procedure (Percent: Double) of object;

  TFractal2 = class(TFrame)
    ImgAll: TImage;
    ShapeZoom: TShape;
    procedure ImgAllMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure ImgAllMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure ImgAllMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
  private
    FRadius: Double;
    FRect: TDoubleRect;
    FSelectX, FSelectY: Integer;
    FOnViewChange: TViewChangeProcedure;
    FOnProgress: TProgressProcedure;
    FShiftComplex: TComplex;

    function RecurseColor(C: TComplex): TColor;
    procedure RepaintFractal;

    procedure SetView(ARect: TDoubleRect);
    procedure SetRadius(ARadius: Double);
    procedure SetShiftComplex(AShiftComplex: TComplex);
  public
    property View: TDoubleRect read FRect write SetView;
    property Radius: Double read FRadius write SetRadius;
    property ShiftComplex: TComplex read FShiftComplex write SetShiftComplex;
    property OnViewChange: TViewChangeProcedure read FOnViewChange write FOnViewChange;
    property OnProgress: TProgressProcedure read FOnProgress write FOnProgress;

    constructor Create(AOwner: TComponent); override;
    procedure Reset;
  end;

implementation

{$R *.DFM}

const
  ColorPalette : Array[1..16] of TColor = (clBlack, clRed, clGreen, clYellow, clBlue, clWhite, clGray, clFuchsia, clTeal, clNavy, clMaroon, clLime, clOlive, clPurple, clSilver, clAqua);

constructor TFractal2.Create(AOwner: TComponent);
begin
  inherited;

  Radius := 5;
  Reset;
end;

function TFractal2.RecurseColor(C: TComplex): TColor;
var
  I: Integer;
  C2: TComplex;
begin
  for I:= 0 to 255 do
  begin
    C2.Re := C.Re * C.Re - C.Img * C.Img + ShiftComplex.Re;
    C2.Img := 2 * C.Re * C.Img + ShiftComplex.Img;
    C.Re := C2.Re;
    C.Img := C2.Img;
    if (C.Re * C.Re + C.Img * C.Img >= FRadius * FRadius) then
    begin
      Result := ColorPalette[I mod 16 + 1];
      Exit;
    end;
  end;
  Result := ColorPalette[256 mod 16 + 1];
end;

procedure TFractal2.RepaintFractal;
  function Trans(X, Y: Double): TComplex;
  begin
    Result.Re := X / ImgAll.Width * (FRect.Right - FRect.Left) + FRect.Left;
    Result.Img := Y / ImgAll.Height * (FRect.Bottom - FRect.Top) + FRect.Top;
  end;
var
  X, Y: Integer;
begin
  
  for X:= 0 to ImgAll.Width - 1 do
  begin
    for Y:= 0 to ImgAll.Height - 1 do
    begin
      ImgAll.Canvas.Pixels[X, Y] := RecurseColor(Trans(X, Y));
    end;
    if Assigned(FOnProgress) then
      FOnProgress(X / (ImgAll.Width - 1));
  end;
end;

procedure TFractal2.ImgAllMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  FSelectX := X;
  FSelectY := Y;
  with ShapeZoom do
  begin
    Visible := true;
    Width := 0;
    Height := 0;
  end;
end;

function Min(A, B: Integer): Integer;
begin
  if A > B then
    Result := B
  else
    Result := A;
end;

function Max(A, B: Integer): Integer;
begin
  if A < B then
    Result := B
  else
    Result := A;
end;

procedure TFractal2.ImgAllMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
begin
  if ShapeZoom.Visible then
  begin
    with ShapeZoom do
    begin
      Left := Min(FSelectX, X);
      Top := Min(FSelectY, Y);
      Width := Abs(FSelectX - X);
      Height := Abs(FSelectY - Y);
    end;
  end;
end;

procedure TFractal2.ImgAllMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var
  NewRect: TDoubleRect;
begin
  ShapeZoom.Visible := false;
  if (Abs(FSelectX - X) < 10) or (Abs(FSelectY - Y) < 10) then
    Exit;
  with NewRect do
  begin
    Left := Min(FSelectX, X) / ImgAll.Width * (FRect.Right - FRect.Left) + FRect.Left;
    Top := Min(FSelectY, Y) / ImgAll.Height * (FRect.Bottom - FRect.Top) + FRect.Top;
    Right := Max(FSelectX, X) / ImgAll.Width * (FRect.Right - FRect.Left) + FRect.Left;
    Bottom := Max(FSelectY, Y) / ImgAll.Height * (FRect.Bottom - FRect.Top) + FRect.Top;
  end;
  View := NewRect;
end;

procedure TFractal2.SetView(ARect: TDoubleRect);
begin
  FRect := ARect;

  if Assigned(FOnViewChange) then
    FOnViewChange(FRect);
  RepaintFractal;
end;

procedure TFractal2.SetRadius(ARadius: Double);
begin
  FRadius := ARadius;
  RepaintFractal;
end;

procedure TFractal2.SetShiftComplex(AShiftComplex: TComplex);
begin
  FShiftComplex := AShiftComplex;
  RepaintFractal;
end;

procedure TFractal2.Reset;
var
  ARect: TDoubleRect;
begin
  with ARect do
  begin
    Left := -Radius;
    Top := Radius;
    Right := Radius;
    Bottom := -Radius;
  end;
  View := ARect;
end;

end.
