unit Fractal3Unit;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, Fractal2Unit, ComplexUnit;

type
  TFractal3Mode = (tf3mRoot, tf3mSpeed);

  TFractal3 = class(TFrame)
    ImgAll: TImage;
    ShapeZoom: TShape;
    procedure ImgAllMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure ImgAllMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure ImgAllMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
  private
    FDegree: Integer;
    FIterTime: Integer;
    FMode: TFractal3Mode;
    FRoot: array of TComplex;
    FRect: TDoubleRect;
    FSelectX, FSelectY: Integer;
    FOnViewChange: TViewChangeProcedure;
    FOnProgress: TProgressProcedure;

    function GetColor(C: TComplex): TColor;
    procedure RepaintFractal;
    function NearRoot(C: TComplex): Integer;
    procedure CalculateRoot;
    function ColorKind(Decision:Integer; Index: Integer): TColor;
    function NextGuess(Comp: TComplex): TComplex;

    procedure SetDegree(ADegree: Integer);
    procedure SetIterTime(AIterTime: Integer);
    procedure SetView(ARect: TDoubleRect);
    procedure SetMode(AMode: TFractal3Mode);

  public
    constructor Create(AOwner: TComponent); override;
    property View: TDoubleRect read FRect write SetView;
    property Degree:Integer read FDegree write SetDegree;
    property IterTime:Integer read FIterTime write SetIterTime;
    property Mode: TFractal3Mode read FMode write SetMode;
    property OnViewChange: TViewChangeProcedure read FOnViewChange write FOnViewChange;
    property OnProgress: TProgressProcedure read FOnProgress write FOnProgress;

    procedure Reset;
  end;

implementation

{$R *.DFM}

const EPSILON = 0.001;

procedure TFractal3.SetView(ARect: TDoubleRect);
begin
  FRect := ARect;
  if Assigned(FOnViewChange) then
    FOnViewChange(FRect);
  RepaintFractal;
end;

procedure TFractal3.SetMode(AMode: TFractal3Mode);
begin
  FMode := AMode;
  RepaintFractal;
end;

constructor TFractal3.Create(AOwner: TComponent);
begin
  inherited;

  FDegree := 3;
  FIterTime := 10;
  FMode := tf3mSpeed;
  CalculateRoot;
  Reset;
end;

procedure TFractal3.CalculateRoot;
var
  I: Integer;
begin
  SetLength(FRoot, Degree);
//The root of X^Degree = 1
  for I:= 0 to Degree-1 do
  begin
    with FRoot[I] do
    begin
      Re := Cos(2.0 * PI / Degree * I);
      Img := Sin(2.0 * PI / Degree * I);
    end;
  end;
end;

procedure TFractal3.SetDegree(ADegree: Integer);
begin
  FDegree := ADegree;
  CalculateRoot;
  RepaintFractal;
end;

procedure TFractal3.SetIterTime(AIterTime: Integer);
begin
  FIterTime := AIterTime;
  RepaintFractal;
end;

function TFractal3.NearRoot(C: TComplex): Integer;
  function DistanceSquare(C1, C2: TComplex): Double;
  var
    Temp: TComplex;
  begin
    Temp := CompSub(C1, C2);
    Result := Temp.Re*Temp.Re + Temp.Img*Temp.Img;
  end;
var
  I: Integer;
begin
  for I:= 0 to Degree-1 do
  begin
    if (DistanceSquare(FRoot[I], C) < EPSILON) then
    begin
      Result := I;
      Exit;
    end;
  end;
  Result := -1;
end;

const
  arrColorBase: array[0..6] of TColor
    = ($00010000, $00000100, $00000001, $00010100, $00000101, $00010001, $00010101);

const
  ColorPalette: array[1..14] of TColor
    = (clRed, clBlue, clYellow, clGreen, clWhite, clFuchsia, clTeal, clNavy, clMaroon, clLime, clOlive, clPurple, clSilver, clAqua);

function TFractal3.ColorKind(Decision:Integer; Index: Integer): TColor;
begin
  if FMode = tf3mRoot then
    Result := arrColorBase[Decision] * (255 - (255 * Index) div FIterTime)
  else
    Result := ColorPalette[1 + Index mod 14];
end;

function TFractal3.NextGuess(Comp: TComplex): TComplex;
  function F(C: TComplex): TComplex;
  begin
    Result := CompExp(C, Degree);
    Result.Re := Result.Re - 1.0;
  end;
  function FDiff(C: TComplex): TComplex;
  begin
    Result := CompExp(C, Degree-1);
    Result.Re := Result.Re * Degree;
    Result.Img := Result.Img * Degree;
  end;
var
  Temp: TComplex;
begin
  Temp := CompDiv(F(Comp), FDiff(Comp));
  Result.Re := Comp.Re - Temp.Re;
  Result.Img := Comp.Img - Temp.Img;
end;

function TFractal3.GetColor(C: TComplex): TColor;
var
  I: Integer;
  C2: TComplex;
  Decision: Integer;
begin
  C2.Re := C.Re;
  C2.Img := C.Img;
  for I:= 0 to FIterTime - 1 do
  begin
    Decision := NearRoot(C2);
    if Decision >= 0 then
    begin
      Result := ColorKind(Decision, I);
      Exit;
    end;

    C2 := NextGuess(C2);
  end;
  Result := clBlack;
end;

procedure TFractal3.RepaintFractal;
  function Trans(X, Y: Double): TComplex;
  begin
    Result.Re := X / ImgAll.Width * (FRect.Right - FRect.Left) + FRect.Left;
    Result.Img := Y / ImgAll.Height * (FRect.Bottom - FRect.Top) + FRect.Top;
  end;
var
  X, Y: Integer;
begin
  ImgAll.Canvas.FillRect(RECT(0, 0, ImgAll.Width, ImgAll.Height));

  for X:= 0 to ImgAll.Width - 1 do
  begin
    for Y:= 0 to ImgAll.Height - 1 do
    begin
      ImgAll.Canvas.Pixels[X, Y] := GetColor(Trans(X, Y));
    end;
    if Assigned(FOnProgress) then
      FOnProgress(X / (ImgAll.Width - 1));
  end;
end;

procedure TFractal3.ImgAllMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  FSelectX := X;
  FSelectY := Y;
  with ShapeZoom do
  begin
    Visible := true;
    Width := 0;
    Height := 0;
  end;
end;

function Min(A, B: Integer): Integer;
begin
  if A > B then
    Result := B
  else
    Result := A;
end;

function Max(A, B: Integer): Integer;
begin
  if A < B then
    Result := B
  else
    Result := A;
end;

procedure TFractal3.ImgAllMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
begin
  if ShapeZoom.Visible then
  begin
    with ShapeZoom do
    begin
      Left := Min(FSelectX, X);
      Top := Min(FSelectY, Y);
      Width := Abs(FSelectX - X);
      Height := Abs(FSelectY - Y);
    end;
  end;
end;

procedure TFractal3.ImgAllMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var
  NewRect: TDoubleRect;
begin
  ShapeZoom.Visible := false;
  if (Abs(FSelectX - X) < 10) or (Abs(FSelectY - Y) < 10) then
    Exit;
  with NewRect do
  begin
    Left := Min(FSelectX, X) / ImgAll.Width * (FRect.Right - FRect.Left) + FRect.Left;
    Top := Min(FSelectY, Y) / ImgAll.Height * (FRect.Bottom - FRect.Top) + FRect.Top;
    Right := Max(FSelectX, X) / ImgAll.Width * (FRect.Right - FRect.Left) + FRect.Left;
    Bottom := Max(FSelectY, Y) / ImgAll.Height * (FRect.Bottom - FRect.Top) + FRect.Top;
  end;
  View := NewRect;
end;

procedure TFractal3.Reset;
var
  ARect: TDoubleRect;
begin
  with ARect do
  begin
    Left := -1.5;
    Top := 1.5;
    Right := 1.5;
    Bottom := -1.5;
  end;
  View := ARect;
end;


end.
